/*
 * Copyright 1999, Be Incorporated.
 * Copyright 2007, Haiku.
 * Distributed under the terms of the MIT License.
 *
 * Authors:
 *		Be Incorporated
 *		Eric Petit <eric.petit@lapsus.org>
 */

#include <KernelExport.h>
#include <PCI.h>
#include <OS.h>
#include <malloc.h>
#include <string.h>
#include "Driver.h"


DeviceData *gPd;
pci_module_info *gPciBus;


/*--------------------------------------------------------------------*/
/* api_version: tells the kernel what revision of the driver API we
 * support */

int32 api_version = B_CUR_DRIVER_API_VERSION;


/*--------------------------------------------------------------------*/
/* init_hardware: returns B_OK if a device is found, B_ERROR
 * otherwise */

status_t
init_hardware()
{
	status_t ret = B_ERROR;
	pci_info pcii;
	int i;

	TRACE("init_hardware\n");

	if (get_module(B_PCI_MODULE_NAME, (module_info **)&gPciBus) != B_OK)
		return B_ERROR;

	for (i = 0; (*gPciBus->get_nth_pci_info)(i, &pcii) == B_OK; i++) {
		if (pcii.vendor_id == PCI_VENDOR_ID_VMWARE &&
			pcii.device_id == PCI_DEVICE_ID_VMWARE_SVGA2) {
			ret = B_OK;
			break;
		}
	}

	put_module(B_PCI_MODULE_NAME);

	TRACE("init_hardware: %ld\n", ret);
	return ret;
}


/*--------------------------------------------------------------------*/
/* init_driver: allocates needed resources */

status_t
init_driver()
{
	status_t ret = B_OK;
	int i;

	TRACE("init_driver\n");

	if (get_module(B_PCI_MODULE_NAME, (module_info **)&gPciBus) != B_OK) {
		ret = B_ERROR;
		goto done;
	}

	if (!(gPd = calloc(1, sizeof(DeviceData)))) {
		put_module(B_PCI_MODULE_NAME);
		ret = B_ERROR;
		goto done;
	}

	/* Remember the PCI information */
	for (i = 0; (*gPciBus->get_nth_pci_info)(i, &gPd->pcii) == B_OK; i++) 
		if (gPd->pcii.vendor_id == PCI_VENDOR_ID_VMWARE &&
			gPd->pcii.device_id == PCI_DEVICE_ID_VMWARE_SVGA2)
			break;

	/* Create a benaphore for exclusive access in OpenHook/FreeHook */
	INIT_BEN(gPd->kernel);

	/* The device name */
	gPd->names[0] = strdup("graphics/vmware");
	gPd->names[1] = NULL;

	/* Usual initializations */
	gPd->isOpen = 0;
	gPd->sharedArea = -1;
	gPd->si = NULL;

done:
	TRACE("init_driver: %ld\n", ret);
	return ret;
}


/*--------------------------------------------------------------------*/
/* publish_devices: returns the list of supported devices */

const char **
publish_devices()
{
	TRACE("publish_devices\n");
	return (const char **)gPd->names;
}


device_hooks *
find_device(const char *name)
{
	TRACE("find_device (%s)\n", name);
	if (gPd->names[0] && !strcmp(gPd->names[0], name))
		return &gGraphicsDeviceHooks;
	return NULL;
}


void uninit_driver()
{
	TRACE("uninit_driver\n");
	free(gPd);
	put_module(B_PCI_MODULE_NAME);
}

